/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
q3master@hendriklipka.de
*/

package de.hendriklipka.q3master;

/**
 * Title:        Quake3Arena MasterBrowser
 * Description:  queries server data
 * NOTE: this code is from gcp (see http://www.hendriklipka.de/java/gcp.html)
 * Copyright:    Copyright (c) 2001 by Hendrik Lipka
 * @author Hendrik Lipka
 * @version 0.1
*/

import java.util.*;
import java.net.*;
import java.io.*;

import de.hendriklipka.util.*;

import org.apache.log4j.*;

public class ServerStatusParser
{
    private Category CAT=Category.getInstance(ServerStatusParser.class.getName());
    private Category LOG_CAT=Category.getInstance("q3master.status");


    protected int _timeOut=2000;
    protected int _tries=3;

    public ServerStatusParser()
    {
        _timeOut=Configuration.getInstance().getServerQueryTimeout();
        _tries=Configuration.getInstance().getServerQueryTries();
    }

    public boolean queryServerStatus(ServerData sd)
    {
        String s=null;
        for (int i=0;i<_tries;i++)
        {
            CAT.info("sending status request to server");
            s=sendARequest("getstatus",sd._serverAddress,sd._serverPort);
            if (null!=s)
                break;
        }
        if (null==s)
            return false;
        CAT.info("got status response");
        Properties serverProps=parseServerInfo1(s);

        s=(String)serverProps.get("PLAYERNUM");
        if (null!=s)
            try
            {
                sd._players=Integer.parseInt(s);
            }
            catch (Exception ex) {}

        s=(String)serverProps.get("sv_maxclients");
        if (null!=s)
            try
            {
                sd._maxPlayers=Integer.parseInt(s);
            }
            catch (Exception ex) {}

        s=(String)serverProps.get("protocol");
        if (null!=s)
            try
            {
                sd._protocol=Integer.parseInt(s);
            }
            catch (Exception ex) {}
        return true;
    }

    // parse "getstatus"
    /*
    \maxclients\8\spawn\0\watervis\0\*version\2.30\*gamedir\fortress\hostname\TF on w
    anderer\maxspectators\5\maxplayer\10\maxplayers\10\s\off\g\on\deathmatch\3\teamp
    lay\31?TF2.5\timelimit\30\fraglimit\100\samelevel\2\*progs\59258\map\2fort5\n\2
    -1 0 "Keel" // ID qport??? "Name"
    */
    public Properties parseServerInfo1(String s)
    {
        if (s.startsWith("statusResponse\n"))
            s=s.substring(15);
        int pos=s.indexOf('\n');
        String pl=null;
        if (-1!=pos)
        {
            pl=s.substring(pos);
            s=s.substring(0,pos);
        }
        Properties props=new Properties();
        if (s.startsWith("\\")) // could be missing
            s=s.substring(1);
        StringTokenizer st=new StringTokenizer(s,"\\");
        try
        {
            while (st.hasMoreElements())
            {
                String name=((String)st.nextElement()).trim();
                String value=((String)st.nextElement()).trim();
                props.put(name,value);
            }
        }
        catch (Exception ex) {} // ignore - maybe we are out of sync...

        // parse existing players
        int players=0;
        if (null!=pl)
        {
            st=new StringTokenizer(pl,"\n");
            while (st.hasMoreElements())
            {
                String line=(String)st.nextElement();
                line=line.trim();
                if (0==line.length())
                    continue;
                players++;
            }
            props.put("PLAYERNUM",Integer.toString(players));
        }
        return props;
    }

    public String sendARequest(String command, InetAddress address, int port)
    {
        String result=null;
        try
        {
            byte[] p=sendPacket(command,address,port);
            int len=p.length;
            StringBuffer sb=new StringBuffer();
            for (int i=0;i<len;i++)
            sb.append((char)p[i]);
            result=resultToString(sb.toString());
        }
        catch (Exception ex)
        { // ignore, but return null
        }
        return result;
    }

    protected byte[] sendPacket(String command,InetAddress IP, int Port) throws Exception
    {
        DatagramSocket ds=null;
        ds=UDPSocketPool.getSocket();
        try
        {
            ds.setSoTimeout(_timeOut);
            if (null==ds)
                return null;

            int cmdLen=command.length();
            byte[] ba=new byte[cmdLen+5];
            ba[0]=(byte)255;
            ba[1]=(byte)255;
            ba[2]=(byte)255;
            ba[3]=(byte)255;
            for (int i=0;i<cmdLen;i++)
            {
                ba[4+i]=(byte)command.charAt(i);
            }
            DatagramPacket dp= new DatagramPacket(ba,ba.length,IP,Port);
            ds.send(dp);
            byte[] buffer=new byte[65507]; // maximum packet size
            DatagramPacket pr=new DatagramPacket(buffer,buffer.length);
            ds.receive(pr);
            if (IP.equals(pr.getAddress())&&
            (Port==pr.getPort()))
            {
                byte[] br=new byte[pr.getLength()];
                System.arraycopy(pr.getData(),0,br,0,pr.getLength());
                return br;
            }
        }
        finally
        {
            UDPSocketPool.freeSocket(ds);
        }
        return null;
    }

    protected String resultToString(String s)
    {
        while ((char)-1==s.charAt(0))
            s=s.substring(1);
        if (s.startsWith("n"))
            s=s.substring(1);
        if (s.startsWith("\\"))
            s=s.substring(1);
        return s;
    }
}